@@ -32,101 +32,28 @@
 (defgroup org-node-context nil "Preview backlink contexts in separate buffer."
   :group 'org-node)
 
-
-;;; Persistence
-
-(defcustom org-node-context-persist-on-disk nil
-  "Whether to sync cached backlink previews to disk.
-The disk location is `org-node-data-dir'.
-
-This allows the context buffer created by \\[org-node-context-raise] to
-show up more instantly, even the first time it renders a given set of
-backlinks.
-
-Here\\='s how you test whether this is noticeable on your system:
-1. Open a file with many subtree nodes
-2. Collapse them all (i.e. S-TAB until you see only the headings)
-3. Have a context buffer open with `org-node-context-follow-mode'
-4. Hold the C-n or <down> key."
-  :type 'boolean
-  :package-version '(org-node . "2.0.0"))
-
-(defvar org-node-context--previews (make-hash-table :test 'equal)
-  "1:N table mapping IDs to seen previews of backlink contexts.
-
-Each preview is a cons cell \(POS-DIFF . TEXT) where POS-DIFF
-corresponds to a link\\='s buffer position relative to that of
-the heading that has said ID, and TEXT is an output of
-`org-node-context--get-preview'.")
-
-(defvar org-node-context--persist-timer (timer-create))
-(defvar org-node-context--last-tbl-state 0)
-(defvar org-node-context--did-init-persist nil)
-
-(defun org-node-context--maybe-init-persistence (&rest _)
-  "Try to restore `org-node-context--previews' from disk.
-Then start occasionally syncing back to disk.
-No-op if user option `org-node-context-persist-on-disk' is nil."
-  (when org-node-context-persist-on-disk
-    (unless org-node-context--did-init-persist
-      (setq org-node-context--did-init-persist t)
-      (cancel-timer org-node-context--persist-timer)
-      (setq org-node-context--persist-timer
-            (run-with-idle-timer 50 t #'org-node-context--persist))
-      ;; Load from disk.
-      (when (file-readable-p (org-node-context--persist-file))
-        (with-temp-buffer
-          (insert-file-contents (org-node-context--persist-file))
-          (let ((data (read (current-buffer))))
-            (when (hash-table-p data)
-              (setq org-node-context--last-tbl-state (hash-table-count data))
-              (setq org-node-context--previews data))))))))
-
-(defun org-node-context--persist-file ()
-  "Return path to file that caches previews between sessions."
-  (mkdir org-node-data-dir t)
-  (file-name-concat org-node-data-dir "org-node-backlink-previews.eld"))
-
-(defun org-node-context--persist ()
-  "Sync all cached previews to disk."
-  (if org-node-context-persist-on-disk
-      ;; Only proceed if table has gained new entries.
-      ;; NOTE: Does not proceed if there are merely new previews in existing
-      ;;       entries, but it's good enough this way.
-      (when (not (eq org-node-context--last-tbl-state
-                     (hash-table-count org-node-context--previews)))
-        (org-node-context--clean-stale-previews)
-        (setq org-node-context--last-tbl-state
-              (hash-table-count org-node-context--previews))
-        (with-temp-file (org-node-context--persist-file)
-          (let ((print-length nil))
-            (prin1 org-node-context--previews (current-buffer)))))
-    (cancel-timer org-node-context--persist-timer)
-    (setq org-node-context--did-init-persist nil)))
-
 (defun org-node-context--clean-stale-previews ()
-  "Clean stale members in table `org-node-context--previews'.
-
-Note that each entry in that table has potentially many previews,
-but when this finds one of them stale, it removes that whole entry."
+  "Clean stale members in table `org-node-context--previews'."
   (let ((valid-positions (make-hash-table :test 'equal)))
-    (maphash
-     (lambda (_ links)
-       (dolist (link links)
-         (push (org-mem-link-pos link)
-               (gethash (org-mem-link-nearby-id link) valid-positions))))
-     org-mem--target<>links)
 
-    (maphash
-     (lambda (id previews)
-       (let ((node (org-mem-entry-by-id id))
-             (valid (gethash id valid-positions)))
-         (or (and node
-                  (cl-loop
-                   for (pos-diff . _text) in previews
-                   always (memq (+ pos-diff (org-mem-entry-pos node)) valid)))
-             (remhash id org-node-context--previews))))
-     org-node-context--previews)))
+    (maphash (lambda (pseudo-id links)
+               (dolist (link links)
+                 (push (org-mem-link-pos link)
+                       (gethash (or (org-mem-link-nearby-id link) pseudo-id)
+                                valid-positions))))
+             org-mem--pseudo-id<>links)
+
+    (maphash (lambda (key previews)
+               (let ((entry (or (org-mem-entry-by-id key)
+                                (org-mem-entry-by-pseudo-id key)))
+                     (valid (gethash key valid-positions)))
+                 (unless (and entry
+                              (cl-loop
+                               for (pos-diff . _text) in previews
+                               always (memq (+ pos-diff (org-mem-entry-pos entry))
+                                            valid)))
+                   (remhash key org-node-context--previews))))
+             org-node-context--previews)))
 
 
 ;;; Early defs
@@ -317,18 +244,22 @@ properties.  Org-mode is enabled, but the org-element cache is not."
   (interactive () org-node-context-mode)
   (unless (derived-mode-p 'org-node-context-mode)
     (error "`org-node-context-visit-thing' called outside context buffer"))
+  ;; Bit magical, but `magit-insert-section' can store a
+  ;; link as the "value" at that section.
   (let* ((value-atpt (oref (magit-current-section) value))
          link-pos
-         (node (if (org-mem-entry-p value-atpt)
-                   ;; Bit magical, but `magit-insert-section' could store the
-                   ;; node as the "value" at that section.
-                   value-atpt
-                 (setq link-pos (org-mem-link-pos value-atpt))
-                 (org-node-context--get-link-origin value-atpt))))
-    (org-node-goto node)
-    (when link-pos
-      (goto-char link-pos)
-      (recenter))))
+         (origin (pcase value-atpt
+                   ((pred org-mem-entry-p) nil)
+                   ((pred org-mem-link-p)
+                    (setq link-pos (org-mem-link-pos value-atpt))
+                    (org-node-context--get-link-origin value-atpt))
+                   ((pred listp) nil))))
+    (if origin
+        (progn
+          (org-node-goto origin)
+          (goto-char link-pos)
+          (recenter))
+      (message "Point not on a backlink snippet"))))
 
 (defun org-node-context-raise-1 ()
   "Either display a context buffer or refresh an already visible one."
@@ -426,6 +357,13 @@ otherwise call the latter."
 
 ;;; Plumbing
 
+(defvar org-node-context--previews (make-hash-table :test 'equal)
+  "Table of preview snippets of backlink contexts.
+
+Each table value is an alist of \((POS-DIFF . TEXT) ...) where POS-DIFF
+corresponds to a link\\='s buffer position relative to that of its
+heading, and TEXT is an output of `org-node-context--get-preview'.")
+
 (defun org-node-context--refresh (&optional buf id from-history-nav)
   "Refresh buffer BUF to show context for node known by ID.
 
@@ -455,14 +393,14 @@ that buffer."
         (erase-buffer)
         (setq header-line-format
               (concat "Context for \"" (org-mem-entry-title node) "\""))
-        (magit-insert-section (org-node-context node)
+        (magit-insert-section (org-node-context :root)
           (when-let* ((links (org-mem-id-links-to-entry node)))
-            (magit-insert-section (org-node-context 'id-links)
+            (magit-insert-section (org-node-context :id-links)
               (magit-insert-heading "ID backlinks:")
               (org-node-context--insert-backlink-sections links)
               (insert "\n")))
           (when-let* ((links (org-mem-roam-reflinks-to-entry node)))
-            (magit-insert-section (org-node-context 'reflinks)
+            (magit-insert-section (org-node-context :reflinks)
               (magit-insert-heading "Ref backlinks:")
               (org-node-context--insert-backlink-sections links)
               (insert "\n"))))
@@ -494,8 +432,7 @@ Actually, if a snippet was previously cached, return the cached version,
 else briefly visit the file, go to the link and call
 `org-node-context--extract-entry-at-point'."
   (let* ((entry (org-node-context--get-link-origin link))
-         ;; TODO: Change to upcoming `org-mem-entry-pseudo-id'
-         (eid (or (org-mem-entry-id entry) entry))
+         (eid (or (org-mem-entry-id entry) (org-mem-entry-pseudo-id entry)))
          (link-pos (org-mem-link-pos link))
          ;; NOTE: `pos-diff' is not necessary in a simple implementation, but
          ;; this level of granularity lets us avoid wiping all cached previews
@@ -553,9 +490,71 @@ Actually return the ancestor ID-node, if there is one and the direct
 entry has no ID."
   (let ((id (org-mem-link-nearby-id link)))
     (or (and id (org-mem-entry-by-id id))
-        ;; TODO: Change to upcoming `org-mem-link-entry'
-        (org-mem-entry-at-pos-in-file (org-mem-link-pos link)
-                                      (org-mem-link-file link)))))
+        (org-mem-link-entry link))))
+
+
+;;; Persistence
+
+(defcustom org-node-context-persist-on-disk nil
+  "Whether to sync cached backlink previews to disk.
+The disk location is `org-node-data-dir'.
+
+This allows the context buffer created by \\[org-node-context-raise] to
+show up more instantly, even the first time it renders a given set of
+backlinks.
+
+Here\\='s how you test whether this is noticeable on your system:
+1. Open a file with many subtree nodes
+2. Collapse them all (i.e. S-TAB until you see only the headings)
+3. Have a context buffer open with `org-node-context-follow-mode'
+4. Hold the C-n or <down> key."
+  :type 'boolean
+  :package-version '(org-node . "2.0.0"))
+
+(defvar org-node-context--persist-timer (timer-create))
+(defvar org-node-context--last-tbl-state 0)
+(defvar org-node-context--did-init-persist nil)
+
+(defun org-node-context--maybe-init-persistence (&rest _)
+  "Try to restore `org-node-context--previews' from disk.
+Then start occasionally syncing back to disk.
+No-op if user option `org-node-context-persist-on-disk' is nil."
+  (when org-node-context-persist-on-disk
+    (unless org-node-context--did-init-persist
+      (setq org-node-context--did-init-persist t)
+      (cancel-timer org-node-context--persist-timer)
+      (setq org-node-context--persist-timer
+            (run-with-idle-timer 50 t #'org-node-context--persist))
+      ;; Load from disk.
+      (when (file-readable-p (org-node-context--persist-file))
+        (with-temp-buffer
+          (insert-file-contents (org-node-context--persist-file))
+          (let ((data (read (current-buffer))))
+            (when (hash-table-p data)
+              (setq org-node-context--last-tbl-state (hash-table-count data))
+              (setq org-node-context--previews data))))))))
+
+(defun org-node-context--persist-file ()
+  "Return path to file that caches previews between sessions."
+  (mkdir org-node-data-dir t)
+  (file-name-concat org-node-data-dir "org-node-backlink-previews.eld"))
+
+(defun org-node-context--persist ()
+  "Sync all cached previews to disk."
+  (if org-node-context-persist-on-disk
+      ;; Only proceed if table has gained new entries.
+      ;; NOTE: Does not proceed if there are merely new previews in existing
+      ;;       entries, but it's good enough this way.
+      (when (not (eq org-node-context--last-tbl-state
+                     (hash-table-count org-node-context--previews)))
+        (org-node-context--clean-stale-previews)
+        (setq org-node-context--last-tbl-state
+              (hash-table-count org-node-context--previews))
+        (with-temp-file (org-node-context--persist-file)
+          (let ((print-length nil))
+            (prin1 org-node-context--previews (current-buffer)))))
+    (cancel-timer org-node-context--persist-timer)
+    (setq org-node-context--did-init-persist nil)))
 
 (provide 'org-node-context)
 
